import numpy as np

__all__ = ["play_mono_audio"]

DEFAULT_SAMPLERATE: int = 44100


def play_mono_audio(audio: np.ndarray, samplerate: int, audio_chunk_size: int = 1024) -> None:
    import pyaudio  # type: ignore[reportMissingImports], pyaudio is optional to install,

    if audio.dtype not in [np.float32, np.float16, np.float64]:
        raise ValueError("Input audio datatype is expected to be a floating point")
    (n_samples,) = audio.shape
    # better to clip then to overflow
    audio = np.clip(audio, -1.0, 1.0)
    # very dumb conversion to PCM16
    pcm_audio = (audio * np.iinfo(np.int16).max).astype(np.int16)

    audio_chunk_size = 1024
    num_chunks = int(np.ceil(n_samples / audio_chunk_size))

    # actual size of each chunk might not be exactly 'audio_chunk_size' but not critical here
    chunks = np.array_split(pcm_audio, num_chunks)

    audio_interface = pyaudio.PyAudio()
    output_stream = audio_interface.open(
        format=audio_interface.get_format_from_width(pcm_audio.dtype.itemsize),
        channels=1,
        rate=samplerate,
        output=True,
    )

    for chunk in chunks:
        output_stream.write(chunk.tobytes())



from collections.abc import Iterable, Mapping
from dataclasses import dataclass, replace
from functools import cached_property
from typing import Self, TypedDict

import jax
from jaxtyping import Array, DTypeLike, PRNGKeyArray
from jinja2 import Template
from tokenizers import Tokenizer

from lalamo.common import ParameterTree, require_tree
from lalamo.modules.common import DummyUnionMember, LalamoModule, register_config_union
from lalamo.sampling import SamplingPolicy, make_policy

from .audio_decoder import TTSAudioDecoder
from .fishaudio.fishaudio_audio_decoding import DescriptAudioCodecConfig
from .fishaudio.fishaudio_text_decoding import FishAudioTextDecoderConfig
from .text_decoder import TTSTextDecoder
from .vocoders import Vocoder, VocoderConfig

__all__ = ["TTSMessage", "TTSMessageProcessor", "TTSMessageProcessorConfig"]

DEFAULT_TTS_SAMPLING_POLICY: SamplingPolicy = make_policy(temperature=0.3, top_p=0.9)
DEFAULT_TTS_REPETITION_PENALTY: float = 1.1


@dataclass(frozen=True)
class VoicePrompt:
    """
    Current class is reserved for future usage of audio prompts
    to condition style of generated audio
    """


@dataclass(frozen=True)
class TTSMessage:
    content: str
    speaker_id: str
    style: str


class TTSRequest(TypedDict):
    messages: list[TTSMessage]


@dataclass(frozen=True)
class TTSMessageProcessorConfig:
    prompt_template: str

    # TODO(peter.glushkov): find a better way to handle opening new-line symbol
    drop_initial_newline: bool = True

    def init(self, tokenizer: Tokenizer) -> "TTSMessageProcessor":
        return TTSMessageProcessor(
            self,
            tokenizer=tokenizer,
        )


@dataclass(frozen=True)
class TTSMessageProcessor:
    config: TTSMessageProcessorConfig
    tokenizer: Tokenizer

    @cached_property
    def prompt_template(self) -> Template:
        return Template(self.config.prompt_template)

    def request_to_dict(
        self,
        messages: Iterable[TTSMessage],
    ) -> TTSRequest:
        return TTSRequest(messages=list(messages))

    def render_request(self, messages: Iterable[TTSMessage]) -> str:
        request_dict = self.request_to_dict(messages)
        prompt_text = self.prompt_template.render({**request_dict})
        if self.config.drop_initial_newline and prompt_text.startswith("\n"):
            prompt_text = prompt_text[1:]
        return prompt_text

    def tokenize_text(self, text: str) -> list[int]:
        return self.tokenizer.encode(text, add_special_tokens=False).ids

    def tokenize_request(self, messages: Iterable[TTSMessage]) -> list[int]:
        rendered = self.render_request(messages)
        return self.tokenize_text(rendered)

    def detokenize(self, tokens: list[int]) -> str:
        return self.tokenizer.decode(tokens, skip_special_tokens=False)


TTSAudioDecoderConfig = DescriptAudioCodecConfig | DummyUnionMember
register_config_union(TTSAudioDecoderConfig)

TTSTextDecoderConfig = FishAudioTextDecoderConfig | DummyUnionMember
register_config_union(TTSTextDecoderConfig)


@dataclass(frozen=True)
class TTSConfig:
    text_decoder_config: TTSTextDecoderConfig
    audio_decoder_config: TTSAudioDecoderConfig
    vocoder_config: VocoderConfig

    activation_precision: DTypeLike

    def empty(self) -> "TTSModel":
        text_decoder = self.text_decoder_config.empty()
        audio_decoder = self.audio_decoder_config.empty()
        vocoder = self.vocoder_config.empty()
        return TTSModel(config=self, text_decoder=text_decoder, audio_decoder=audio_decoder, vocoder=vocoder)

    def random_init(self, key: PRNGKeyArray) -> "TTSModel":
        key_text, key_audio = jax.random.split(key)
        text_decoder = self.text_decoder_config.random_init(key=key_text)
        audio_decoder = self.audio_decoder_config.random_init(key=key_audio)
        vocoder = self.vocoder_config.empty()
        return TTSModel(config=self, text_decoder=text_decoder, audio_decoder=audio_decoder, vocoder=vocoder)


class TTSModel(LalamoModule[TTSConfig]):
    text_decoder: TTSTextDecoder
    audio_decoder: TTSAudioDecoder
    vocoder: Vocoder

    @property
    def activation_precision(self) -> DTypeLike:
        return TTSConfig.activation_precision

    def export_weights(self) -> ParameterTree[Array]:
        return {}

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            text_decoder=self.text_decoder.import_weights(require_tree(weights["text_decoder"])),
            audio_decoder=self.audio_decoder.import_weights(require_tree(weights["audio_decoder"])),
        )



from typing import overload

import jax.numpy as jnp
import torch
from bidict import bidict
from jaxtyping import Array

__all__ = ["jax_to_torch", "torch_to_jax"]


class DTypeConvert:
    """Bidirectional JAX <-> PyTorch dtype converter."""

    _map = bidict(
        {
            "float16": torch.float16,
            "float32": torch.float32,
            "float64": torch.float64,
            "bfloat16": torch.bfloat16,
            "int8": torch.int8,
            "int16": torch.int16,
            "int32": torch.int32,
            "int64": torch.int64,
            "uint8": torch.uint8,
            "bool": torch.bool,
            "complex64": torch.complex64,
            "complex128": torch.complex128,
        },
    )

    @classmethod
    def to_torch(cls, in_type: str | jnp.dtype) -> torch.dtype:
        match in_type:
            case str():
                return cls._map[in_type]
            case _:
                return cls._map[jnp.dtype(in_type).name]

    @classmethod
    def to_jax(cls, in_type: str | torch.dtype) -> jnp.dtype:
        match in_type:
            case str():
                return jnp.dtype(in_type)
            case torch.dtype():
                return jnp.dtype(cls._map.inverse[in_type])


@torch.no_grad()
def _torch_to_jax_bfloat16(tensor: torch.Tensor) -> Array:
    if tensor.dtype != torch.bfloat16:
        raise ValueError("Trying to convert non-bfloat16 tensor to bfloat16")
    intermediate_tensor = tensor.view(torch.uint16)
    return jnp.array(intermediate_tensor).view("bfloat16")


@overload
def torch_to_jax(value: torch.Tensor) -> Array: ...
@overload
def torch_to_jax(value: str | torch.dtype) -> jnp.dtype: ...
def torch_to_jax(value: torch.Tensor | str | torch.dtype) -> Array | jnp.dtype:  # type: ignore[return-value]
    if isinstance(value, torch.Tensor):
        value = value.detach().cpu()
        if value.dtype == torch.bfloat16:
            return _torch_to_jax_bfloat16(value)
        return jnp.array(value.numpy())
    if isinstance(value, str | torch.dtype):
        return DTypeConvert.to_jax(value)


@overload
def jax_to_torch(value: Array) -> torch.Tensor: ...
@overload
def jax_to_torch(value: str | jnp.dtype) -> torch.dtype: ...
def jax_to_torch(value: Array | str | jnp.dtype) -> torch.Tensor | torch.dtype:
    if isinstance(value, str | jnp.dtype):
        return DTypeConvert.to_torch(value)
    if isinstance(value, Array):
        from torch.utils import dlpack as _dlpack

        if value.dtype == jnp.bfloat16:
            intermediate_array = value.view(jnp.uint16)
            return _dlpack.from_dlpack(intermediate_array).view(torch.bfloat16)
        return _dlpack.from_dlpack(value)
